import re
import os
import time
import json
from pathlib import Path
from hashlib import md5
from jsonmodels.errors import ValidationError
from description_harvester.models.description import Component

#Functions to make DACS dates from timestamps and ISO dates
def stamp2DACS(timestamp):
	calendar = {"01": "January", "02": "February", "03": "March", "04": "April", "05": "May", "06": "June", "07": "July", "08": "August", "09": "September", "10": "October", "11": "November", "12": "December"}
	stamp = timestamp[:8]
	year = stamp[:4]
	month = stamp[4:6]
	day = stamp[-2:]
	normal = year + "-" + month + "-" + day
	if day.startswith("0"):
		day = day[-1:]
	dacs = year + " " + calendar[month] + " " + day
	return dacs, normal
	
def iso2DACS(normalDate):
	calendar = {'01': 'January', '02': 'February', '03': 'March', '04': 'April', '05': 'May', '06': 'June', '07': 'July', '08': 'August', '09': 'September', '10': 'October', '11': 'November', '12': 'December'}
	if "-" in normalDate:
		if normalDate.count('-') == 1:
			year = normalDate.split("-")[0]
			month = normalDate.split("-")[1]
			displayDate = year + " " + calendar[month]
		else:
			year = normalDate.split("-")[0]
			month = normalDate.split("-")[1]
			day = normalDate.split("-")[2]
			if day.startswith("0"):
				displayDay = day[1:]
			else:
				displayDay = day
			displayDate = year + " " + calendar[month] + " " + displayDay
	else:
		displayDate = normalDate
	return displayDate


def extract_years(date_string):
    # Initialize an empty set to store unique years
    years = set()

    # Regular expression to match year ranges, e.g., "1988-2001"
    range_pattern = r'(\d{4})-(\d{4})'
    # Regular expression to match individual years in text (e.g., "1988 July 29")
    individual_year_pattern = r'(\d{4})'

    # Find all year ranges and individual years in the date string
    ranges = re.findall(range_pattern, date_string)
    individual_years = re.findall(individual_year_pattern, date_string)

    # Add years from the ranges to the set (inclusive)
    for start, end in ranges:
        years.update(range(int(start), int(end) + 1))  # `range` is inclusive of start but exclusive of end

    # Add individual years directly to the set
    years.update(int(year) for year in individual_years)

    # Return the sorted list of unique years
    return sorted(years)

# Caching to disk functionality

CACHE_DIR = Path.home() / ".description_harvester" / "cache"
CACHE_DIR.mkdir(parents=True, exist_ok=True)

def write2disk(object, collection_id):
	# takes a jsonmodel object and writes it to disk for testing

	test_path = Path.home() / ".description_harvester"
	test_path.mkdir(parents=True, exist_ok=True)
	out_path = test_path / f"{collection_id}.json"

	with open(out_path, "w") as file:
		json.dump(object.to_dict(), file, indent=4)

def get_cache_key(identifier):
    # Handles special characters and URIs cleanly
    return md5(str(identifier).encode()).hexdigest() + ".json"

def component_from_dict(data):
    """
    Recursively rebuild a Component instance (and its nested children) from a dict.
    """
    # First create the base Component without components
    components_data = data.pop('components', [])
    component = Component(**data)

    # Now recursively load any nested components
    for child in components_data:
        component.components.append(component_from_dict(child))

    return component

def save_to_cache(identifier, data):
    """
    Save the given jsonmodel data to a cache file.
    """
    key = get_cache_key(identifier)
    
    # Convert `data` (a jsonmodel object) to a dictionary using `to_dict()`
    if hasattr(data, 'to_dict'):
        data = data.to_dict()  # Convert the jsonmodel object to a dictionary
    
    # Ensure the cache directory exists
    CACHE_DIR.mkdir(parents=True, exist_ok=True)
    
    # Save the data along with a timestamp
    with open(CACHE_DIR / key, "w") as f:
        json.dump({
            "timestamp": int(time.time()),
            "data": data
        }, f)

def load_from_cache(identifier, max_age_seconds=86400):
    key = get_cache_key(identifier)
    path = CACHE_DIR / key
    
    if path.exists():
        with open(path, "r") as f:
            cached = json.load(f)
            age = time.time() - cached["timestamp"]
            if age < max_age_seconds:
                try:
                    print (f"\tReading from cache: {path}")
                    return component_from_dict(cached["data"])
                except Exception as e:
                    print(f"\tError rebuilding model: {e}")
                    return None
    return None



from os import listdir, makedirs
from os.path import basename, dirname, exists, isfile, join
from importlib.machinery import SourceFileLoader

from abc import ABC, abstractmethod

class Plugin(ABC):
    """Derivative - abstract base class and registry for reading digital object data from a number of sources.
    Since this is likely to require implementation-specific logic, these are designed as customizable plugins
    This class serves two purposes.  Firstly, it serves as an abstract base class, defining
    common methods and properties required to read digital object data into the description model.
    Secondly, it serves as a registry of such formats, with the goal of allowing new formats to
    be implemented in a "plug-in" fashion.  Any class subclassing this one, with the expected
    properties and a class variable "plugin_name" will be registered on this class and thus usable
    by the software."""

    # Registry of plugins, key = cls.plugin_name, value = cls
    registry = {}

    def __init_subclass__(cls, **kwargs):
        """Enforce plugin descriptive attributes on subclasses, register them"""
        plugin_attrs = ["plugin_name"]
        for attr in plugin_attrs:
            if not hasattr(cls, attr):
                raise RuntimeError("Plugin subclass must have `" + attr + "` attribute")

        super().__init_subclass__(**kwargs)
        __class__.registry[cls.plugin_name] = cls


    #@abstractmethod
    def custom_repository(self, resource):
        # Returns a repository name as a string based on custom logic from data in resource API object
        pass

    #@abstractmethod
    def read_data(self, href):
        # parses digital object data from plugin managing digital objects and dao-level metadata
        pass


def import_plugins(additional_dirs=None):
    if not additional_dirs:
        additional_dirs = []

    dirs = [join(dirname(__file__), "plugins"), *additional_dirs]


    for plugin_dir in dirs:
        if not exists(plugin_dir):
            continue
        for filename in listdir(plugin_dir):
            module = basename(filename)[:-3]
            full_path = join(plugin_dir, filename)
            # skip if not a normal, non underscored file ending in .py
            if module.startswith("_") or not isfile(full_path) or filename[-3:] != ".py":
                continue
            SourceFileLoader(module, full_path).load_module()



import time
import yaml
import pysolr
import argparse
from pathlib import Path
from datetime import datetime
from .version import __version__
from .configurator import Config
from .utils import write2disk, save_to_cache, load_from_cache
from description_harvester.outputs.arclight import Arclight
from description_harvester.inputs.aspace import ArchivesSpace

config = Config()

def parse_args(args=None):
    parser = argparse.ArgumentParser(description='Description_harvester manages archival description.')
    parser.add_argument('--version', action='version', version=f'description_harvester {__version__}')
    parser.add_argument('-v', '--verbose', action="store_true")
    parser.add_argument('--id', nargs="+")
    parser.add_argument('--uri', nargs="+")
    parser.add_argument('--delete', nargs="+")
    parser.add_argument('--updated', action="store_true")
    parser.add_argument('--new', action="store_true")
    parser.add_argument('--hour', action="store_true")
    parser.add_argument('--today', action="store_true")
    parser.add_argument('--solr_url', nargs=1)
    parser.add_argument('--core', nargs=1)
    parser.add_argument('--repo')
    parser.add_argument('--repo_id', type=int)
    #parser.add_argument('--ead', default=False, action="store_true", help='Optionally write to a EAD file(s).')
    
    return parser.parse_args(args) if args else parser.parse_args()

def post_record(arclight, record, verbose=False):
    solr_doc = arclight.convert(record)
    arclight.post(solr_doc)
    print(f"\tIndexed {record.id}")

def get_time_since(args):
    if args.updated:
        return config.last_query
    if args.hour:
        return str(int(time.time()) - 3600)
    if args.today:
        return str(int(time.time()) - 86400)
    return None

def index_record(arclight, aspace, collection_id, use_uri=False, verbose=False):
    loader = aspace.read_uri if use_uri else aspace.read
    record = load_from_cache(collection_id, config.cache_expiration)
    if not record:
        record = loader(collection_id)
        if record:
            save_to_cache(collection_id, record)
    if record:
        post_record(arclight, record, verbose)

def handle_deletions(solr_url, solr_core, collection_ids):
    solr = pysolr.Solr(f"{solr_url}/{solr_core}", always_commit=True)
    solr.ping()
    for collection_id in collection_ids:
        solr.delete(id=collection_id.replace(".", "-"))
        print(f"\tDeleted {collection_id}")

def harvest(args=None):
    args = parse_args(args)
    start_time = time.time()
    print(f"\n------------------------------\nRan at: {datetime.fromtimestamp(start_time)}")

    if not (args.id or args.new or args.updated or args.uri or args.hour or args.today or args.delete):
        print("No action requested, need a collection ID or --updated or --new")
        return

    solr_url = args.solr_url[0] if args.solr_url else config.solr_url
    solr_core = args.core[0] if args.core else config.solr_core

    if args.delete:
        handle_deletions(solr_url, solr_core, args.delete)
        return

    repository_name = Config.read_repositories(args.repo) if args.repo else None
    repo_id = str(args.repo_id) if args.repo_id else '2'
    aspace = ArchivesSpace(repository_id=repo_id, verbose=args.verbose)
    arclight = Arclight(f"{solr_url}/{solr_core}", repository_name)

    time_since = get_time_since(args)
    if time_since is not None:
        for uri in aspace.read_since(time_since):
            print (uri)
            index_record(arclight, aspace, uri, use_uri=True, verbose=args.verbose)

    if args.new:
        solr = pysolr.Solr(f"{solr_url}/{solr_core}", always_commit=True)
        solr.ping()
        for cid in aspace.all_resource_ids():
            results = solr.search(f"id:{cid.replace('.', '-')}", rows=1, **{"fl": "id"})
            if results.hits == 0:
                index_record(arclight, aspace, cid, verbose=args.verbose)
            else:
                print(f"\tSkipping {cid} (already exists)")

    if args.id:
        for cid in args.id:
            index_record(arclight, aspace, cid, verbose=args.verbose)

    if args.uri:
        for uri in args.uri:
            index_record(arclight, aspace, uri, use_uri=True, verbose=args.verbose)

    end_time = time.time()
    config.last_query = str(int(end_time))
    with open(Path.home() / ".description_harvester.yml", "w") as f:
        yaml.dump(config.__dict__, f)

    print(f"Stored last run time as: {datetime.utcfromtimestamp(end_time)}")
    print(f"Execution time: {end_time - start_time:.2f} seconds")

	


def filter_empty_fields(obj):
    """
    Recursively filters out empty fields from a given object or list.
    """
    if isinstance(obj, dict):
        # If obj is a dictionary, filter out empty values
        return {k: filter_empty_fields(v) for k, v in obj.items() if v not in [None, '', [], {}]}
    elif isinstance(obj, list):
        # If obj is a list, recursively filter each item
        return [filter_empty_fields(v) for v in obj if v not in [None, '', [], {}]]
    else:
        # If it's a basic value (not a list or dict), return it
        return obj
        